#include "utils.h"
#include "layout.h"
#include "wingedmonkey.h"
#include "atomic.h"
#include "cmdeng.h"

#include "sock.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>

unsigned int color_nb, color_sb;


xstate XSTATE;
Monitor* MONITOR;

static int running=False;
static int restack_hold=0;

static int ignore_keys=0;

char* hook_unmanage=0;
char* hook_manage=0;

static void (*handler[LASTEvent]) (XEvent *) = {
	[ButtonPress] = event_mouseclick,
	[ClientMessage] = event_client_message,
	[ConfigureRequest] = event_configure_request,
	[ConfigureNotify] = event_configure_notify,
	[DestroyNotify] = event_destroy_notify,
	[FocusIn] = event_client_focusin,
	[FocusOut] = event_client_focusout,
	[KeyPress] = event_keypress,
	[MappingNotify] = event_mappingnotify,
	[MapRequest] = event_client_maprequest,
	[PropertyNotify] = event_client_propertynotify,
	[UnmapNotify] = event_unmapnotify
};

#define ClkRootWin         0x01
#define ClkClientWin       0x02

#define ClkMaskFocus       0x10
#define ClkMaskUnfocus     0x20

#define ClkClientFocused   (ClkClientWin|ClkMaskFocus)
#define ClkClientUnfocused (ClkClientWin|ClkMaskUnfocus)
#define ClkClientAlways    (ClkClientWin|ClkMaskFocus|ClkMaskUnfocus)

static const Button buttons[] = {
  { ClkClientUnfocused,    0,            Button1,      eventfun_mousefocus },
  { ClkClientAlways,       MODKEY,       Button1,      eventfun_mousemove },
  { ClkClientAlways,       MODKEY,       Button3,      eventfun_mouseresize },
  { ClkRootWin,            0,            Button4,      eventfun_scrollup },
  { ClkRootWin,            0,            Button5,      eventfun_scrolldn },
};

//shift means move window
static const Key keys[] = {
  { MODKEY,                 XK_Tab,        user_alttab, CycleAll},
  { LINKEY,                 XK_Tab,        user_alttab, CycleTiles},

  { LINKEY|SHIFTKEY,        XK_Tab,        user_sendtonexttile, 0},
  { LINKEY,                 XK_Page_Down,  user_nextintile, 1},
  { LINKEY,                 XK_Page_Up,    user_nextintile, -1},
  { MODKEY,                 XK_F4,         user_kill, 0},
  { LINKEY,                 XK_k,          user_kill, 0},
  { LINKEY,                 XK_9,          user_kill, 9},
  { 0,                      XK_Escape,     user_killtransient, 0},
  { LINKEY,                 XK_q,          user_quit, 0},
  { LINKEY,                 XK_m,          user_client_tog_flrel, FloatMax},
  { LINKEY,                 XK_f,          user_client_tog_flrel, FloatFull},
  { LINKEY,                 XK_t,          user_client_tog_strel, StckTop},
  { LINKEY,                 XK_b,          user_client_tog_strel, StckBottom},
  //layers
  { LINKEY,                 XK_Home,       user_togglelayer, 0},
  { LINKEY,                 XK_F1,         user_togglelayer, 1},
  { LINKEY,                 XK_F2,         user_togglelayer, 2},
  { LINKEY,                 XK_F3,         user_togglelayer, 3},
  { LINKEY,                 XK_F4,         user_togglelayer, 4},
  { LINKEY,                 XK_F5,         user_togglelayer, 5},
  { LINKEY,                 XK_F6,         user_togglelayer, 6},
  { LINKEY,                 XK_F7,         user_togglelayer, 7},
  { LINKEY,                 XK_F8,         user_togglelayer, 8},
  { LINKEY,                 XK_F9,         user_togglelayer, 9},
  //send to layers
  { LINKEY|SHIFTKEY,        XK_Home,       user_sendtolayer, 0},
  { LINKEY|SHIFTKEY,        XK_F1,         user_sendtolayer, 1},
  { LINKEY|SHIFTKEY,        XK_F2,         user_sendtolayer, 2},
  { LINKEY|SHIFTKEY,        XK_F3,         user_sendtolayer, 3},
  { LINKEY|SHIFTKEY,        XK_F4,         user_sendtolayer, 4},
  { LINKEY|SHIFTKEY,        XK_F5,         user_sendtolayer, 5},
  { LINKEY|SHIFTKEY,        XK_F6,         user_sendtolayer, 6},
  { LINKEY|SHIFTKEY,        XK_F7,         user_sendtolayer, 7},
  { LINKEY|SHIFTKEY,        XK_F8,         user_sendtolayer, 8},
  { LINKEY|SHIFTKEY,        XK_F9,         user_sendtolayer, 9},

  //desktops
  { CTRLKEY,                XK_F1,         user_desktopswitch, 1},
  { CTRLKEY,                XK_F2,         user_desktopswitch, 2},
  { CTRLKEY,                XK_F3,         user_desktopswitch, 3},
  { CTRLKEY,                XK_F4,         user_desktopswitch, 4},
  { CTRLKEY,                XK_F5,         user_desktopswitch, 5},
  { CTRLKEY,                XK_F6,         user_desktopswitch, 6},
  { CTRLKEY,                XK_F7,         user_desktopswitch, 7},
  { CTRLKEY,                XK_F8,         user_desktopswitch, 8},
  { CTRLKEY,                XK_F9,         user_desktopswitch, 9},
  //switch desktop
  { LINKEY,                 XK_Left,       user_desktopnext, -1},
  { LINKEY,                 XK_Right,      user_desktopnext, 1},

  { MODKEY,                 XK_Return,     user_terminal, 0}
};

static unsigned int numlockmask = 0;
static unsigned int lastmask = 0; //mask of the last key event

static Window stupid;


///////////////////////////////////////////
static int (*xerrorxlib)(Display*, XErrorEvent*);
static int otherwm=False;

static int xerrordummy(Display* dpy, XErrorEvent* ee) {
  return 0;
}
static int xerrorstart(Display* dpy, XErrorEvent* ee) {
  otherwm = True;
  return -1;
}

static int xerror(Display* dpy, XErrorEvent* ee) {
  if(ee->error_code == BadWindow
     || (ee->request_code == X_SetInputFocus && ee->error_code == BadMatch)
     || (ee->request_code == X_ConfigureWindow && ee->error_code == BadMatch)
     || (ee->request_code == X_GrabButton && ee->error_code == BadAccess)
     || (ee->request_code == X_GrabKey && ee->error_code == BadAccess))
    return 0;
  say("wingedmonkey: fatal error: request code=%d, error code=%d\n",ee->request_code, ee->error_code);
  return xerrorxlib(XSTATE.dpy, ee);
}

static unsigned long getcolor(const char* colstr) {
  Colormap cmap = DefaultColormap(XSTATE.dpy, XSTATE.screen);
  XColor color;
  if(!XAllocNamedColor(XSTATE.dpy, cmap, colstr, &color, &color))
    die("error, cannot allocate color '%s'\n", colstr);
  return color.pixel;
}

static void sigterm(int sig){
  say("caught signal %d\n",sig);
  running=0;
  signal(SIGTERM,SIG_DFL);
  signal(SIGINT,SIG_DFL);
  signal(SIGQUIT,SIG_DFL);
  signal(SIGHUP,SIG_DFL);
}


static void sigchld(int ignored){
  if(signal(SIGCHLD, sigchld) == SIG_ERR){
    die("Can't install SIGCHLD handler\n");
  }
  while(0 < waitpid(-1, NULL, WNOHANG));
}


////////////////////////////////////////


unsigned int xorg_keycodetomodmask(unsigned int keycode) {
  XModifierKeymap* modmap = XGetModifierMapping(XSTATE.dpy);
  if (!modmap)return 0;

  // search through modmap for this keycode
  for (int mod=0; mod < 8; mod++) {
    for (int key=0; key < modmap->max_keypermod; ++key) {
      // modifiermap is an array with 8 sets of keycodes
      // each max_keypermod long, but in a linear array.
      if (modmap->modifiermap[modmap->max_keypermod*mod + key] == keycode) {
	XFreeModifiermap(modmap);
	return (1<<mod);//modlist[mod].mask;
      }
    }
  }
  XFreeModifiermap(modmap);
  // no luck
  return 0;
}

static void xorg_stacktransients(Client* parent, Window* ptr, int* index){
  Client* child=parent->children;
  while(child){
    if(child->children){
      xorg_stacktransients(child,ptr,index);
    }
    ptr[(*index)++]=child->win;
    child=child->chld_next;
  }
}

int xorg_restack(){
  if(restack_hold)return 0;

  Window* wins=malloc(MONITOR->clientcount*sizeof(Window));
  int count=0;
  if(MONITOR->focus && MONITOR->focus->float_state==FloatFull){
    wins[count++]=MONITOR->focus->win;
  }
  for(int stackl=StckTop;stackl>=StckBottom;stackl--){
    for(Layer* l=MONITOR->sfirstlayer;l;l=l->snext){
      if(l->stacktype!=stackl)continue;
      for(int stacklayer=StckDock;stacklayer>=StckDesktop;stacklayer--){
	for(Client* c=MONITOR->firstfocus;c;c=c->fnext){
	  if(c->layer!=l)continue;
	  if(c->stacktype!=stacklayer)continue;
	  if(c==MONITOR->focus && MONITOR->focus->float_state==FloatFull)continue;
	  if(c->owner)continue;
	  if(c->children)xorg_stacktransients(c,wins,&count);
	  wins[count++]=c->win;
	}
      }
    }
  }
  if(!count){free(wins);return 0;}
  XRaiseWindow(XSTATE.dpy,wins[0]);
  XRestackWindows(XSTATE.dpy,wins,count);
  atomic_setprop_windowa(XSTATE.root,net_atoms[NetClientListStacking],wins,count);
  free(wins);
  XSync(XSTATE.dpy, False);
  XEvent ev;
  while(XCheckMaskEvent(XSTATE.dpy, EnterWindowMask, &ev));//clear events
  return 0;
}


int xorg_recalc(){
  XGrabServer(XSTATE.dpy);//hold everything
  for(Layer* l=MONITOR->firstlayer;l;l=l->next){
    for(Tile* t=l->firsttile;t;t=t->next){
      tile_recalc(t);
    }
  }
  for(Client* c=MONITOR->firstclient;c;c=c->next){
    client_reshape(c,0);//if fullscreen, it will reshape it
    xclient_moveresize(c);
  }
  XUngrabServer(XSTATE.dpy);
  return 0;
}

void xorg_checkprotocols(Client* c){
  int n;
  Atom* protocols;
  c->protocols=0;
  if(XGetWMProtocols(XSTATE.dpy, c->win, &protocols, &n)) {
    for(int i = 0; i < n; i++){
      if(protocols[i] == wm_atoms[WMDelete]){
	c->protocols|=PRTCL_DELETE;
	continue;
      }
      if(protocols[i] == wm_atoms[WMTakeFocus]){
	c->protocols|=PRTCL_FOCUS;
	continue;
      }
    }
    XFree(protocols);
  }
}

void xorg_getsizehints(Client *c) {
  long msize;
  XSizeHints size;

  // size is uninitialized, ensure that size.flags aren't used
  if(!XGetWMNormalHints(XSTATE.dpy, c->win, &size, &msize))size.flags=PSize;

  if(size.flags & PBaseSize) {
    c->basew = size.base_width;
    c->baseh = size.base_height;
  }else if(size.flags & PMinSize) {
    c->basew = size.min_width;
    c->baseh = size.min_height;
  }
  else {
    c->basew = c->baseh = 0;
  }

  if(size.flags & PResizeInc) {
    c->incw = size.width_inc;
    c->inch = size.height_inc;
  }else {
    c->incw = c->inch = 0;
  }

  if(size.flags & PMaxSize) {
    c->maxw = size.max_width;
    c->maxh = size.max_height;
  }else{
    c->maxw = c->maxh = 0;
  }

  if(size.flags & PMinSize) {
    c->minw = size.min_width;
    c->minh = size.min_height;
  }else if(size.flags & PBaseSize) {
    c->minw = size.base_width;
    c->minh = size.base_height;
  }else{
    c->minw = c->minh = 0;
  }

  if(size.flags & PAspect) {
    c->mina = ((float)size.min_aspect.y) / size.min_aspect.x;
    c->maxa = ((float)size.max_aspect.x) / size.max_aspect.y;
  }else{
    c->maxa = c->mina = 0.0;
  }

  c->hints&=~MY_HINT_FIXED;
  if(c->minw==c->maxw && c->minh==c->maxh && c->minw>0 && c->minh>0){
    c->hints|=MY_HINT_FIXED;
  }
}

void xorg_getstrings(Client* c){
  strcpy(c->name, "unknown");
  strcpy(c->role, "default");
  strcpy(c->classN, "none");
  strcpy(c->classC, "none");
  atomic_getstring(c->win, wm_atoms[WMName], c->name);
  atomic_getstring(c->win, net_atoms[NetWMName], c->name);
  atomic_getstring(c->win, wm_atoms[WMWindowRole], c->role);
  //atomic_getstring(c->win, wm_atoms[WMClass], c->class);
  XClassHint* xch=XAllocClassHint();
  XGetClassHint(XSTATE.dpy,c->win,xch);
  if(xch->res_name){
    strcpy(c->classN,xch->res_name);
    XFree(xch->res_name);
  }
  if(xch->res_class){
    strcpy(c->classC,xch->res_class);
    XFree(xch->res_class);
  }
  XFree(xch);
}

/*
void xorg_clearurgent(Client* c) {
  XWMHints* wmh;

  c->hints &= ~(MY_HINT_URGNT);
  if(!(wmh = XGetWMHints(XSTATE.dpy, c->win)))return;
  wmh->flags &= ~XUrgencyHint;
  XSetWMHints(XSTATE.dpy, c->win, wmh);
  XFree(wmh);
}

void xorg_updatewmhints(Client* c) {
  if(!c)return;
  XWMHints* wmh=0;

  c->hints &= ~(MY_HINT_URGNT);
  if((wmh = XGetWMHints(XSTATE.dpy, c->win))) {
    if((c == MONITOR->focus) && (wmh->flags & XUrgencyHint)) {
      wmh->flags &= ~XUrgencyHint;
      XSetWMHints(XSTATE.dpy, c->win, wmh);
    }else{
      if(wmh->flags & XUrgencyHint){
	c->hints|=MY_HINT_URGNT;
	XSetWindowBorder(XSTATE.dpy, c->win, color_ub);
      }
    }
    XFree(wmh);
  }
}
*/

void xorg_refreshnetwmhints(Client* c, int clearskip){
  if(!c)return;
  if(!c->tile)return;
  int nitems;
  Atom* atoms=atomic_getprop_atoma(c->win,net_atoms[NetWMState],&nitems);

  int changed=0;//if we need to refresh

  //these we will send to the server
  Atom* newatoms=(Atom*)malloc((4+nitems)*sizeof(Atom));
  int newcount=0;

  int is_fullscreen=c->float_state==FloatFull;
  int is_max=c->float_state==FloatMax || c->tile->force_maximize;
  int is_hidden=!c->visible;
  int was_fullscreen=0;
  int was_max=0;
  int was_hidden=0;
  for(unsigned int i=0;i<nitems;i++){
    if(atoms[i]==net_atoms[NetWMFullscreen]){
      was_fullscreen=1;
    }else if(atoms[i]==net_atoms[NetWMMaximizedVert]){
      was_max+=2;
    }else if(atoms[i]==net_atoms[NetWMMaximizedHorz]){
      was_max--;
    }else if(atoms[i]==net_atoms[NetWMHidden]){
      was_hidden=1;
    }else if(atoms[i]==net_atoms[NetWMSkipTaskbar] && clearskip){
      //skip it
    }else{//leave them alone
      newatoms[newcount++]=atoms[i];
    }
  }
  if(clearskip==2)newatoms[newcount++]=net_atoms[NetWMSkipTaskbar];
  if(was_fullscreen!=is_fullscreen)changed=1;
  if(was_max!=is_max)changed=1;
  if(was_hidden!=is_hidden)changed=1;
  if(changed){
    if(is_fullscreen)newatoms[newcount++]=net_atoms[NetWMFullscreen];
    if(is_max){
      newatoms[newcount++]=net_atoms[NetWMMaximizedHorz];
      newatoms[newcount++]=net_atoms[NetWMMaximizedVert];
    }
    if(is_hidden){
      newatoms[newcount++]=net_atoms[NetWMHidden];
    }
    atomic_setprop_atoma(c->win,net_atoms[NetWMState],newatoms,newcount);
    //refreshed
  }

  free(newatoms);
  XFree(atoms);
}

int xorg_checkotherwm() {
  otherwm = False;
  xerrorxlib = XSetErrorHandler(xerrorstart);
  // this causes an error if some other window manager is running
  XSelectInput(XSTATE.dpy, DefaultRootWindow(XSTATE.dpy), SubstructureRedirectMask);
  XSync(XSTATE.dpy, False);
  XSetErrorHandler(xerror);
  XSync(XSTATE.dpy, False);
  return otherwm;
}

int xorg_getrootptr(int* x, int* y) {
  int di;
  unsigned int dui;
  Window dummy;
  return XQueryPointer(XSTATE.dpy, XSTATE.root, &dummy, &dummy, x, y, &di, &di, &dui);
}

void xorg_refreshclientlist(){
  Window* wins=(Window*)malloc(MONITOR->clientcount*sizeof(Window));
  int i=0;
  for(Client* c=MONITOR->firstclient;c;c=c->next){
    wins[i++]=c->win;
  }
  atomic_setprop_windowa(XSTATE.root,net_atoms[NetClientList],wins,i);
  free(wins);
}

void xorg_refreshfocus(){
  Client* focusable=0;
  for(Client* c=MONITOR->firstfocus;c;c=c->fnext){
    if(!c->visible)continue;
    if(c->stacktype>StckDesktop && c->stacktype<StckDock && !c->tile->skip_focus && !(c->hints&MY_HINT_NOFCS)){
      focusable=c;
      break;
    }
  }
  xclient_activate(focusable);
}

void xorg_acknowledgetype(Client* c){
  int nitems;
  Atom* atoms=atomic_getprop_atoma(c->win,net_atoms[NetWMWindowType],&nitems);
  if(nitems==0)return;
  Atom a=atoms[0];
  c->hints|=MY_HINT_TYPES;//it has a type
  if(a==net_atoms[NetWMWindowTypeDesktop]){
    c->stacktype=StckDesktop;
  }else if(a==net_atoms[NetWMWindowTypeDock]){
    c->stacktype=StckDock;
    c->hints|=MY_HINT_NOBRD;
  }else if(a==net_atoms[NetWMWindowTypeToolbar]){
    c->hints|=MY_HINT_FLOAT;
  }else if(a==net_atoms[NetWMWindowTypeMenu]){
    c->hints|=MY_HINT_FLOAT;
  }else if(a==net_atoms[NetWMWindowTypeUtility]){
    c->hints|=MY_HINT_FLOAT;
  }else if(a==net_atoms[NetWMWindowTypeSplash]){
    c->hints|=MY_HINT_FLOAT;
    c->hints|=MY_HINT_NOBRD;
  }else if(a==net_atoms[NetWMWindowTypeDialog]){
    c->hints|=MY_HINT_FLOAT;
    c->hints|=MY_HINT_DIALG;
  }else if(a==net_atoms[NetWMWindowTypeNormal]){
    //do nothing
  }
}

void xorg_refreshclientdesktop(Client* c){
  if(!MONITOR->desktop){
    atomic_setprop_long(c->win,net_atoms[NetWMDesktop],0);
    return;
  }
  int index=layer_to_id(MONITOR,c->layer);
  if(index<0)return;
  long desktop=-2;
  if(MONITOR->desktop->layout[index]=='X'){//if in current
    desktop=MONITOR->desktop->id;
  }else
  if(MONITOR->desktop->layout[index]=='+'){//if everywhere and visible
    desktop=MONITOR->desktop->id;
  }else
  if(MONITOR->desktop->layout[index]=='*'){//if everywhere
    desktop=-2;
  }else{
    for(Dsktp* d=MONITOR->firstdesktop;d;d=d->next){
      if(d->layout[index]=='X'){//it is part of some other desktop
	desktop=d->id;
	break;
      }
    }
  }
  if(desktop!=-3){
    atomic_setprop_long(c->win,net_atoms[NetWMDesktop],desktop);
  }else{
    XDeleteProperty(XSTATE.dpy,c->win,net_atoms[NetWMDesktop]);
  }
  if(desktop==-2){
    xorg_refreshnetwmhints(c,2);
  }else{
    xorg_refreshnetwmhints(c,1);
  }
}

void xorg_refreshalldesktop(){
  for(Client* c=MONITOR->firstclient;c;c=c->next){
    xorg_refreshclientdesktop(c);
  }
}

void xorg_desktopsupdate(){
  atomic_setprop_long(XSTATE.root,net_atoms[NetNumberOfDesktops],MONITOR->desktopcount);
  if(!MONITOR->desktopcount)return;
  char* dnames[MONITOR->desktopcount];
  for(Dsktp* d=MONITOR->firstdesktop;d;d=d->next){
    dnames[d->id]=d->name;
  }
  atomic_setstrings(XSTATE.root,net_atoms[NetDesktopNames],dnames,MONITOR->desktopcount);
}

void xorg_grabkeys(){
  numlockmask=xorg_keycodetomodmask(XKeysymToKeycode(XSTATE.dpy, XK_Num_Lock));
  unsigned int modifiers[] = { 0, LockMask, numlockmask, numlockmask|LockMask };
  KeyCode code;

  int len=ignore_keys?2:LENGTH(keys);

  XUngrabKey(XSTATE.dpy, AnyKey, AnyModifier, XSTATE.root);
  for(int i = 0; i < len; i++) {
    if((code = XKeysymToKeycode(XSTATE.dpy, keys[i].keysym))){
      for(int j = 0; j < LENGTH(modifiers); j++){
	XGrabKey(XSTATE.dpy, code,
		 keys[i].mod | modifiers[j], XSTATE.root,
		 True, GrabModeSync, GrabModeSync);
      }
    }
  }
}

//always only grab mapped buttons, not all of them
void xorg_grabbuttons(Client* c, int focused) {
  numlockmask=xorg_keycodetomodmask(XKeysymToKeycode(XSTATE.dpy, XK_Num_Lock));
  unsigned int modifiers[] = { 0, LockMask, numlockmask, numlockmask|LockMask };
  XUngrabButton(XSTATE.dpy, AnyButton, AnyModifier, c->win);
  unsigned int mask=focused?ClkMaskFocus:ClkMaskUnfocus;
  for(int i = 0; i < LENGTH(buttons);i++){
    if((buttons[i].click & ClkClientWin) && (buttons[i].click & mask) ){
      for(int j = 0; j < LENGTH(modifiers);j++){
	XGrabButton(XSTATE.dpy, buttons[i].button,
		    buttons[i].mask | modifiers[j],
		    c->win, True, BUTTONMASK,
		    GrabModeSync, GrabModeSync, None, None);
      }
    }
  }
}



//////////////////////////////////////////////////////

int xclient_manage(Client* c){
  if(!c)return 1;
  client_add(MONITOR,c);//add to the global list
  xorg_getstrings(c);//get strings ready
  int steal=1;
  Rule* r=client_find_rule(MONITOR,c);
  if(r && r->no_steal){
    say("nosteal found\n");
    steal=0;
  }
  if(steal){
    client_focus_add_front(c);
  }else{
    client_focus_add_second(c);
  }
  Window transient=None;
  Client* transc=0;
  if(XGetTransientForHint(XSTATE.dpy,c->win,&transient)){
    transc=win_to_client(MONITOR,transient);
  }
  if(transc){
    client_addtransient(transc,c);
    client_stack_add_front(transc->tile,c);
  }else{//if not transient, find a suitable space for it
    if(client_stack_apply_rules(MONITOR,c)){
      client_stack_add_to_monitor(MONITOR,c);
    }
  }
  client_post_apply_rules(MONITOR,c);//apply the rules that match

  //gethints
  xorg_getsizehints(c);
  xorg_checkprotocols(c);
  xorg_acknowledgetype(c);
  Atom actions[4]={net_atoms[NetWMActionFullscreen],
		   net_atoms[NetWMActionMaximizeVert],
		   net_atoms[NetWMActionMaximizeHorz],
		   net_atoms[NetWMActionClose]
  };
  atomic_setprop_atoma(c->win,net_atoms[NetWMAllowedActions],actions,4);

  //apply size rules!
  client_rehint(c,0);
  rect_findbest(c,c->tile);
  client_reshape(c,0);

  xclient_configure(c, CWX|CWY|CWWidth|CWHeight|CWBorderWidth);
  XSetWindowBorder(XSTATE.dpy, c->win, color_nb);

  //trap the events from this window
  XSelectInput(XSTATE.dpy, c->win, FocusChangeMask|PropertyChangeMask|StructureNotifyMask);
  xorg_grabbuttons(c,False);

  if(c->layer->visible){
    xclient_show(c);
    if(steal)xclient_activate(c);
  }
  xorg_restack();
  xorg_refreshnetwmhints(c,0);

  XAddToSaveSet(XSTATE.dpy,c->win);
  xorg_refreshclientlist();
  xorg_refreshclientdesktop(c);

  if(hook_manage){
    char mywin[16];
    sprintf(mywin,"0x%x",(uint)c->win);
    if(fork()==0){
      if(XSTATE.dpy)
	close(ConnectionNumber(XSTATE.dpy));
      setsid();
      execlp(hook_manage,hook_manage,mywin,(char*)0);
      say("wingedmonkey: failed to execute hook\n");
      exit(0);
    }
  }

  return 0;
}

//remove client from management list
int xclient_unmanage(Client* c){
  XGrabServer(XSTATE.dpy);
  if(c==MONITOR->focus)xclient_focus(0);
  xclient_hide(c);
  long data[] = { WithdrawnState, None };
  atomic_setprop_longa(c->win,wm_atoms[WMState],data,2);
  client_stack_remove(c);
  if(c->owner){
    client_remtransient(c);
  }else if(c->children){
    Client* tr=c->children;
    while(tr){//make transients independent
      Client* next=tr->chld_next;
      tr->owner=0;
      tr->chld_next=0;
      tr=next;
    }
  }
  client_focus_remove(c);
  client_remove(c);
  xorg_refreshclientlist();
  xorg_refreshfocus();
  XUngrabServer(XSTATE.dpy);

  if(hook_unmanage){
    char mywin[16];
    sprintf(mywin,"0x%x",(uint)c->win);
    if(fork()==0){
      if(XSTATE.dpy)
	close(ConnectionNumber(XSTATE.dpy));
      setsid();
      execlp(hook_unmanage,hook_unmanage,mywin,c->name,c->role,c->classN,c->classC,(char*)0);
      say("wingedmonkey: failed to execute hook\n");
      exit(0);
    }
  }

  free(c);

  return 0;
}

int xclient_show(Client* c){
  if(!c)return 1;
  if(!c->tile)return 1;
  if(c->visible)return 0;
  c->visible=1;
  XMapWindow(XSTATE.dpy,c->win);
  if(c->holdmove)xclient_moveresize(c);//resize if not resized yet
  long data[] = { NormalState, None };
  atomic_setprop_longa(c->win,wm_atoms[WMState],data,2);

  //xorg_setwstate(c->win, NormalState);
  xorg_refreshnetwmhints(c,0);
  return 0;
}

int xclient_hide(Client* c){
  if(!c)return 1;
  if(!c->tile)return 1;
  if(!c->visible)return 0;
  c->visible=0;
  if(c==MONITOR->focus)xclient_focus(0);//unfocus if hidden

  XGrabServer(XSTATE.dpy);

  c->recentunmapserial=NextRequest(XSTATE.dpy);
  //  say("performing unmap #%d\n",c->recentunmapserial);
  XUnmapWindow(XSTATE.dpy,c->win);

  XUngrabServer(XSTATE.dpy);

  long data[] = { IconicState, None };
  atomic_setprop_longa(c->win,wm_atoms[WMState],data,2);

  xorg_refreshnetwmhints(c,0);
  return 0;
}

int xclient_focus(Client* c){
  if(MONITOR->focus==c){
    if(!c)return 0;
    //check if it is really already set
    Window w=0;
    int rev;
    XGetInputFocus(XSTATE.dpy,&w,&rev);
    if(w==c->win)return 0;
  }
  if(c && !c->visible)return 1;//this is fail
  if(c && c->hints&MY_HINT_NOFCS)return 1;
  if(c){
    /*
    if(c->hints&MY_HINT_URGNT){//clear urgent flag
      xorg_clearurgent(c);
    }
    */
    //if(c->protocols&PRTCL_FOCUS){//suggest focus
      //  atomic_sendmsg(c->win,wm_atoms[WMTakeFocus],CurrentTime);
      //}else{//force focus
      XSetInputFocus(XSTATE.dpy, c->win, RevertToPointerRoot, CurrentTime);
      //}
  }else{//unfocus
    XSetInputFocus(XSTATE.dpy, XSTATE.root, RevertToPointerRoot, CurrentTime);
  }
  if(MONITOR->focus){//undo the coloring on previous focus
    XSetWindowBorder(XSTATE.dpy, MONITOR->focus->win, color_nb);
  }
  MONITOR->focus=c;
  if(MONITOR->focus){//refresh coloring of next focus
    XSetWindowBorder(XSTATE.dpy, MONITOR->focus->win, color_sb);
  }
  return 0;
}

int xclient_activate(Client* c){
  xclient_focus(c);
  xclient_raise(c);
  return 0;
}

int xclient_raise(Client* c){
  if(!c)return 1;
  if(!c->layer)return 1;
  if(!c->visible)return 1;//don't raise stuff that is not there
  if(c && c->owner)xclient_raise(c->owner);//recursive raisement
  client_focus_to_front(c);
  xlayer_raise(c->layer);
  xorg_restack();
  return 0;
}

int xclient_moveresize(Client* c){
  if(!c->visible){
    c->holdmove=1;
    return 1;
  }
  c->holdmove=0;

  XMoveResizeWindow(XSTATE.dpy, c->win, c->xx, c->yy,c->ww-2*c->bw,c->hh-2*c->bw);

  return 0;
}

int xclient_configure(Client* c, unsigned int mask){
  if(!c)return 1;
  XWindowChanges wc;
  wc.border_width=c->bw;
  wc.x=c->xx;
  wc.y=c->yy;
  wc.width=c->ww-2*c->bw;
  wc.height=c->hh-2*c->bw;
  XConfigureWindow(XSTATE.dpy, c->win, mask, &wc);
  return 0;
}


int xclient_changelayer(Client* c, Layer* new){
  if(!c)return 1;
  if(!new)return 1;
  if(c->layer==new)return 0;
  if(!new->prev){
    client_stack_add_front(new->firsttile,c);
  }else{
    if(client_stack_add_to_layer(new,c,AddlvlSticky)){
      if(client_stack_add_to_layer(new,c,AddlvlEmpty)){
	if(client_stack_add_to_layer(new,c,AddlvlFree)){
	  if(client_stack_add_to_layer(new,c,AddlvlOther)){
	    return 1;
	  }
	}
      }
    }
  }
  rect_findbest(c,c->tile);
  client_reshape(c,0);
  restack_hold=1;

  xclient_moveresize(c);
  xclient_configure(c,CWBorderWidth);
  if(new->visible)xclient_show(c);
  else xclient_hide(c);
  xorg_refreshfocus();
  restack_hold=0;
  xorg_restack();
  xorg_refreshclientdesktop(c);
  return 0;
}

int xclient_kill(Client* c) {
  if(c->protocols&PRTCL_DELETE) {
    atomic_sendmsg(c->win,wm_atoms[WMDelete],CurrentTime);
  } else {
    XGrabServer(XSTATE.dpy);
    XSetErrorHandler(xerrordummy);
    XSetCloseDownMode(XSTATE.dpy, DestroyAll);
    XKillClient(XSTATE.dpy, c->win);
    XSync(XSTATE.dpy, False);
    XSetErrorHandler(xerror);
    XUngrabServer(XSTATE.dpy);
  }
  return 0;
}

int xclient_murder(Client* c, int sig){
  if(!c)return 1;
  long pid=atomic_getprop_long(c->win,net_atoms[NetWMPid]);
  say("killing %d with signal %d\n",pid,sig);
  if(pid>0){
    kill((pid_t)pid,sig);
  }
  return 0;
}

void user_killtransient(int arg){
  Client* c=MONITOR->focus;
  if(c && c->owner){
    xclient_kill(c);
  }else{
    //fall through
    XAllowEvents(XSTATE.dpy,ReplayKeyboard,CurrentTime);
  }
}

int xclient_stackrel(Client* c, int stackrel){
  if(!c)return 1;
  if(c->stacktype == stackrel)return 1;
  c->stacktype=stackrel;
  xorg_restack();
  return 0;
}

int xclient_floatrel(Client* c, int floatrel){
  if(!c)return 1;
  if(c->float_state == floatrel)return 1;
  int oldstate=c->float_state;
  c->float_state=floatrel;
  client_reshape(c,0);
  xclient_moveresize(c);
  xclient_configure(c,CWBorderWidth);
  if((floatrel==FloatFull && c->sprev==0) || oldstate==FloatFull)xorg_restack();
  xorg_refreshnetwmhints(c,0);
  return 0;
}

//does not refresh focus, as it is an actual part of the focusing
int xlayer_raise(Layer* l){
  if(!l->sprev && l->visible)return 0;
  if(l!=MONITOR->sfirstlayer && MONITOR->sfirstlayer->focusorhide){
    xlayer_visible(MONITOR->sfirstlayer,0);
  }
  layer_stack_to_front(l);
  //xorg_restack();
  return 0;
}


int xlayer_visible(Layer* lay, int visible){
  if(lay->visible==visible)return 0;
  lay->visible=visible;
  if(visible){
    for(Tile* t=lay->firsttile;t;t=t->next){
      for(Client* c=t->firstclient;c;c=c->snext){
	xclient_show(c);
      }
    }
  }else{
    layer_stack_to_back(lay);//if invisible, hide
    if(MONITOR->focus && lay==MONITOR->focus->layer)xclient_focus(0);
    for(Tile* t=lay->firsttile;t;t=t->next){
      for(Client* c=t->firstclient;c;c=c->snext){
	xclient_hide(c);
      }
    }
  }
  return 0;
}

int xlayer_applylayout(char* string, int force){
  char* c=string;
  char* c0=0;
  if(MONITOR->desktop){
    c0=MONITOR->desktop->layout;
  }else{
    c0=string;
  }

  for(Layer* l=MONITOR->firstlayer;l&&*c&&*c0;l=l->next,c++,c0++){
    if(*c=='*')continue;
    if(*c=='+')continue;
    //if forced, then apply everything, otherwise respect desktop
    if(!force && (*c0=='X' || *c0=='0'))continue;
    if(*c=='0')xlayer_visible(l,0);
    else if(*c=='X'){
      xlayer_visible(l,1);
      if(l->focusorhide){
	xlayer_bringfront(l);
      }
    }
  }
  xorg_refreshfocus();
  return 0;
}

//////////////////////////////////////////////////////

int xorg_scan(){
  unsigned int num;
  Window d1, d2;
  Window* wins = NULL;
  XWindowAttributes wa;

  if(XQueryTree(XSTATE.dpy, XSTATE.root, &d1, &d2, &wins, &num)) {
    for(unsigned int i = 0; i < num; i++) {
      if(!XGetWindowAttributes(XSTATE.dpy, wins[i], &wa)
	 || wa.override_redirect || XGetTransientForHint(XSTATE.dpy, wins[i], &d1))
	continue;
      if(wa.map_state == IsViewable || atomic_getprop_long(wins[i],wm_atoms[WMState]) == IconicState){
	Client* c=new_client(wins[i]);
	c->x=wa.x;
	c->y=wa.y;
	c->w=wa.width;
	c->h=wa.height;
	xclient_manage(c);
      }
    }
    for(unsigned int i = 0; i < num; i++) { // now the transients
      if(!XGetWindowAttributes(XSTATE.dpy, wins[i], &wa))
	continue;
      if(wa.override_redirect)continue;
      if(XGetTransientForHint(XSTATE.dpy, wins[i], &d1)
	 && (wa.map_state == IsViewable || atomic_getprop_long(wins[i],wm_atoms[WMState]) == IconicState)){
	Client* c=new_client(wins[i]);
	c->x=wa.x;
	c->y=wa.y;
	c->w=wa.width;
	c->h=wa.height;
	xclient_manage(c);
      }
    }
    if(wins)
      XFree(wins);
  }

  xorg_refreshclientlist();
  xorg_refreshfocus();

  //create ridiculously stupid empty window that noone cares about
  //some stupid panels check this shit
  stupid=XCreateWindow(XSTATE.dpy,XSTATE.root,-1,-1,1,1,0,CopyFromParent,CopyFromParent,CopyFromParent,0,0);
  atomic_setprop_window(stupid,net_atoms[NetSupportingWMCheck],stupid);
  atomic_setprop_window(XSTATE.root,net_atoms[NetSupportingWMCheck],stupid);
  char* wmname="wingedmonkey";
  atomic_setstrings(stupid,net_atoms[NetWMName],&wmname,1);
  return 0;
}

int xorg_dispose(){
  if(hook_manage){free(hook_manage);hook_manage=0;}
  if(hook_unmanage){free(hook_unmanage);hook_unmanage=0;}
  xclient_focus(0);
  Client* next=0;
  for(Client* c=MONITOR->firstclient;c;c=next){
    next=c->next;
    xclient_kill(c);
    say("killing client %s[0x%x]\n",c->name,c->win);
    xclient_unmanage(c);
  }
  XUngrabKey(XSTATE.dpy, AnyKey, AnyModifier, XSTATE.root);
  XUngrabButton(XSTATE.dpy, AnyButton, AnyModifier, XSTATE.root);
  XEvent ev;
  while(XCheckMaskEvent(XSTATE.dpy, 0xffffffff, &ev));//clear events

  XFreeCursor(XSTATE.dpy, XSTATE.cursor[CurNormal]);
  XFreeCursor(XSTATE.dpy, XSTATE.cursor[CurResize]);
  XFreeCursor(XSTATE.dpy, XSTATE.cursor[CurMove]);
  //recursively remove all layers?
  free(MONITOR);
  XSync(XSTATE.dpy, False);
  XSetInputFocus(XSTATE.dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
  XDestroyWindow(XSTATE.dpy,stupid);
  return 0;
}

int xorg_setup(){
  // init atoms
  for(int i=0;i<WMLast;i++){
    wm_atoms[i]=XInternAtom(XSTATE.dpy, wm_strings[i], False);
  }
  for(int i=0;i<NetLast;i++){
    net_atoms[i]=XInternAtom(XSTATE.dpy, net_strings[i], False);
  }

  color_nb=getcolor("#cccccc");
  color_sb=getcolor("#3377ff");

  // clean up any zombies immediately
  sigchld(0);

  // init screen
  XSTATE.screen = DefaultScreen(XSTATE.dpy);
  XSTATE.root = RootWindow(XSTATE.dpy, XSTATE.screen);
  int mw = DisplayWidth(XSTATE.dpy, XSTATE.screen);
  int mh = DisplayHeight(XSTATE.dpy, XSTATE.screen);
  MONITOR=new_monitor(mw,mh);


  //update destkop size hints
  long data[2]={MONITOR->mw,MONITOR->mh};
  atomic_setprop_longa(XSTATE.root,net_atoms[NetDesktopGeometry],data,2);
  data[0]=data[1]=0;
  atomic_setprop_longa(XSTATE.root,net_atoms[NetDesktopViewport],data,2);
  atomic_setprop_window(XSTATE.root,net_atoms[NetActiveWindow],None);


  // init cursors
  XSTATE.cursor[CurNormal] = XCreateFontCursor(XSTATE.dpy, XC_left_ptr);
  XSTATE.cursor[CurResize] = XCreateFontCursor(XSTATE.dpy, XC_sizing);
  XSTATE.cursor[CurMove]   = XCreateFontCursor(XSTATE.dpy, XC_fleur);
  // EWMH support per view
  atomic_setprop_atoma(XSTATE.root,net_atoms[NetSupported],&net_atoms[NetSupported],NetLast);

  // desktops

  Layer* l0=new_layer("default");
  layer_add(MONITOR,l0);
  layer_stack_add_front(MONITOR,l0);
  Tile* t00=new_tile_full("default-tile",0,0,0,0,0,0,1,1);
  t00->sticky=StickSlippery;
  tile_add(l0,t00);
  xlayer_visible(l0,1);

  char* rc=getenv("WINGMON_RC");
  if(!rc)rc=".wingedmonkeyrc";
  FILE* f=fopen(rc,"r");
  if(!f){
    say("config file missing\n");
  }else{
    char buffer[BUFSIZE];
    while(fgets(buffer,BUFSIZE,f)){
      cmd(buffer);
    }
    fclose(f);
  }

  // select for events
  XSetWindowAttributes wa;
  wa.cursor = XSTATE.cursor[CurNormal];
  wa.event_mask = SubstructureRedirectMask|SubstructureNotifyMask
    |ButtonPressMask|EnterWindowMask|LeaveWindowMask
    |StructureNotifyMask|PropertyChangeMask;
  XChangeWindowAttributes(XSTATE.dpy, XSTATE.root, CWEventMask|CWCursor, &wa);
  XSelectInput(XSTATE.dpy, XSTATE.root, wa.event_mask);
  xorg_grabkeys();
  return 0;
}

static void watcher_sock_kill(){
  char* sockname=getenv("WINGMON_SOCK");
  if(!sockname)sockname=".wingedmonkey.sock";
  struct stat filestatus;
  if(stat(sockname,&filestatus)==0){
    if(S_ISSOCK(filestatus.st_mode)){
      unlink(sockname);
    }
  }
}

static int watcher_sock_init(){
  char* sockname=getenv("WINGMON_SOCK");
  if(!sockname)sockname=".wingedmonkey.sock";

  int socketfd=-1;
  struct stat filestatus;
  if(stat(sockname,&filestatus)==0){
    if(S_ISSOCK(filestatus.st_mode)){
      unlink(sockname);
    }else{
      say("file exists and is not a socket, things will probably fail\n");
      return -1;
    }
  }
  socketfd=socket(AF_UNIX,SOCK_STREAM,0);
  fcntl(socketfd, F_SETFD, FD_CLOEXEC);
  fcntl(socketfd, F_SETFL, O_NONBLOCK|fcntl(socketfd,F_GETFL));
  if(socketfd<0){
    say("could not open socket\n");
    return -1;
  }
  struct sockaddr_un serv_addr;
  serv_addr.sun_family=AF_UNIX;
  strcpy(serv_addr.sun_path,sockname);
  if(bind(socketfd,(struct sockaddr*)&serv_addr,sizeof(struct sockaddr_un))){
    say("could not bind stream socket\n");
    return -1;
  }
  listen(socketfd,1);
  say("listening on the socket '%s'\n",sockname);
  return socketfd;
}

int xorg_run() {
  char commandbuffer[BUFSIZE];

  XSync(XSTATE.dpy, False);
  running=True;

  //define and initialize the file descriptors
  int socket_server=watcher_sock_init();
  int socket_client=-1;
  int socket_xorg=ConnectionNumber(XSTATE.dpy);
  say("sockets: server=%d, xorg=%d\n",socket_server,socket_xorg);

  //client state
  int left_to_read=-1;
  char* bufpos=0;

  struct pollfd fdlist[2];
  fdlist[0].fd=socket_xorg;
  fdlist[0].events=POLLIN;
  fdlist[1].fd=socket_server;
  fdlist[1].events=POLLIN;

#define POLLFAILMASK (POLLERR|POLLHUP|POLLNVAL)

  //interruption stops the loop and cleans up
  signal(SIGTERM,sigterm);
  signal(SIGINT,sigterm);
  signal(SIGQUIT,sigterm);
  signal(SIGHUP,sigterm);


  while(running){
    //poll the sockets (only for input)
    int ret=poll(fdlist,2,-1);
    if(ret==-1){
      if(errno==EINTR)continue;//signal caught
      perror("poll returned -1, error: ");
      break;//fail (errno is set)
    }
    //xorgness
    if(fdlist[0].revents&POLLFAILMASK){
      say("xorg socket failed\n");
      break;
    }
    if(fdlist[0].revents&POLLIN){//event has come
      XEvent ev;
      while(XPending(XSTATE.dpy)){
	if(XNextEvent(XSTATE.dpy,&ev)){
	  say("xnextevent failed\n");
	  break;//fail
	}
	if(handler[ev.type])handler[ev.type](&ev);//execute event
      }
    }
    //sockness
    if(socket_client<0){//no open connection, waiting for it
      if(fdlist[1].revents&POLLIN){//there is something to do
	socket_client=accept(socket_server,0,0);
	if(socket_client==-1){
	  continue;
	}
	fcntl(socket_client,F_SETFD,FD_CLOEXEC);
	left_to_read=-1;//reset the leftover message size
	bufpos=commandbuffer;//reset the buffer position
	fdlist[1].fd=socket_client;//from now on, client socket is important
      }
    }else{
      if(fdlist[1].revents&POLLFAILMASK){//socket got closed
	say("socket got huped\n");
	close(socket_client);
	socket_client=-1;
	left_to_read=-1;
	bufpos=commandbuffer;
	fdlist[1].fd=socket_server;
	continue;
      }
      if(fdlist[1].revents&POLLIN){//handle the reading
	int retval=readmessagepart(socket_client,&bufpos,BUFSIZE,&left_to_read);
	if(retval==0 && left_to_read==0){
	  opensocket=socket_client;//setup the socket cmd command will use to report back
	  cmd(commandbuffer);//execute the command
	  opensocket=-1;
	}
	if((retval==0 && left_to_read==0) || retval==-1){
	  close(socket_client);
	  socket_client=-1;
	  left_to_read=-1;
	  bufpos=commandbuffer;
	  fdlist[1].fd=socket_server;
	}
      }
    }
  }
  //cleanup
  if(socket_client)close(socket_client);
  if(socket_server)close(socket_server);
  watcher_sock_kill();
  return 0;
}

void event_mouseclick(XEvent* e) {
  unsigned int click = 0x00;
  if(e->type!=ButtonPress){//discard release events
    XAllowEvents(XSTATE.dpy,ReplayPointer,e->xbutton.time);
    return;
  }
  XButtonPressedEvent* ev = &e->xbutton;
  Client* c=win_to_client(MONITOR,ev->window);
  if(c){
    click=ClkClientWin;
  }else{
    if(ev->window==XSTATE.root)click=ClkRootWin;
  }

  for(int i = 0; i < LENGTH(buttons); i++){
    if((click & buttons[i].click) &&
       buttons[i].func &&
       buttons[i].button == ev->button &&
       CLEANMASK(buttons[i].mask)==CLEANMASK(ev->state)){
      buttons[i].func(c);
      break;
    }
  }
  XAllowEvents(XSTATE.dpy,ReplayPointer,e->xbutton.time);
}

void eventfun_scrollup(Client* c){
  if(c)return;
  user_desktopnext(-1);
  XUngrabPointer(XSTATE.dpy,CurrentTime);
}

void eventfun_scrolldn(Client* c){
  if(c)return;
  user_desktopnext(1);
  XUngrabPointer(XSTATE.dpy,CurrentTime);
}

void eventfun_mousefocus(Client* c) {
  if(!c)return;
  xclient_activate(c);
}

void eventfun_mousemove(Client* c) {
  if(!c)return;
  if(c->float_state==FloatFull)return;
  if(XGrabPointer(XSTATE.dpy, XSTATE.root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync, None, XSTATE.cursor[CurMove], CurrentTime) != GrabSuccess)return;

  XEvent ev;
  int x0, y0;
  if(!xorg_getrootptr(&x0, &y0))return;
  Client copy=*c;
  do {
    XMaskEvent(XSTATE.dpy, MOUSEMASK|SubstructureRedirectMask, &ev);
    switch (ev.type) {
      case ConfigureRequest:
      case MapRequest:
	handler[ev.type](&ev);
	break;
      case MotionNotify:{
	int x=ev.xmotion.x_root;
	int y=ev.xmotion.y_root;
	client_move(c,x,y,x0,y0,&copy);
	int changetile=0;
	if(copy.tile!=c->tile){
	  changetile=1;
	  client_move_to_tile(copy.tile,c);
	}
	xclient_moveresize(c);
	if(changetile){
	  xclient_configure(c,CWBorderWidth);
	  xorg_restack();
	}
	break;
      }
	//future: escape to revert
    }
  } while(ev.type != ButtonRelease);
  XUngrabPointer(XSTATE.dpy, CurrentTime);
  xorg_refreshfocus();
}

void eventfun_mouseresize(Client* c) {
  if(!c)return;
  if(c->float_state!=FloatFloat)return;
  if(c->hints&MY_HINT_FIXED)return;
  if(XGrabPointer(XSTATE.dpy, XSTATE.root, False, MOUSEMASK, GrabModeAsync, GrabModeAsync, None, XSTATE.cursor[CurResize], CurrentTime) != GrabSuccess)return;
  XEvent ev;
  int x0, y0;
  if(!xorg_getrootptr(&x0, &y0))return;
  Client copy=*c;
  do {
    XMaskEvent(XSTATE.dpy, MOUSEMASK|SubstructureRedirectMask, &ev);
    switch (ev.type) {
      case ConfigureRequest:
      case MapRequest:
	handler[ev.type](&ev);
	break;
      case MotionNotify:{
	int x=ev.xmotion.x_root;
	int y=ev.xmotion.y_root;
	client_resize(c,x,y,x0,y0,&copy);
	xclient_moveresize(c);
	break;
      }
    }
  } while(ev.type != ButtonRelease);
  XUngrabPointer(XSTATE.dpy, CurrentTime);
}

void event_keypress(XEvent* e) {
  XKeyEvent* ev = &e->xkey;
  lastmask = CLEANMASK(ev->state);
  //deprecation
  //KeySym keysym = XKeycodeToKeysym(XSTATE.dpy, (KeyCode)ev->keycode, 0);
  //idx = group_index * key_width + shift_level
  KeySym keysym = XkbKeycodeToKeysym(XSTATE.dpy, (KeyCode)ev->keycode, 0,0);
  for(int i = 0; i < LENGTH(keys); i++){
    if(keysym == keys[i].keysym &&
       CLEANMASK(keys[i].mod) == CLEANMASK(ev->state) &&
       keys[i].func){
      keys[i].func(keys[i].arg);
      break;
    }
  }
  //XAllowEvents(XSTATE.dpy,ReplayKeyboard,ev->time);
  XUngrabKeyboard(XSTATE.dpy,CurrentTime);
  lastmask = 0;
}

void event_unmapnotify(XEvent* e){
  //say("event: UnmapNotify\n");
  XUnmapEvent* ev = &e->xunmap;
  Client* c=win_to_client(MONITOR,ev->window);
  if(!c)return;
  //  say("unmapnotify #%d\n",ev->serial);
  /*
  if(c->visible){
    XWindowAttributes wa;
    XGetWindowAttributes(XSTATE.dpy, ev->window, &wa);
    if(wa.map_state != IsUnmapped){
      //if it is mapped, then?
      xclient_unmanage(c);
    }
  }3
  */

  if(c->visible){
    //    say("unmap notify for visible client 0x%x [%s]\n",c->win,c->name);
    if(c->recentunmapserial==ev->serial){
      //      say("false alarm\n");
      return;
    }
    xclient_unmanage(c);//visible means we didn't call it
  }
  /*
  if(c){
    xclient_hide(c);//this is incompatible with our philosophy, breaks everything
  }
  */
}

void event_destroy_notify(XEvent* e){
  XDestroyWindowEvent* ev = &e->xdestroywindow;
  Client* c=win_to_client(MONITOR,ev->window);
  if(!c)return;
  xclient_unmanage(c);
}

void event_client_focusin(XEvent* e) {
  XFocusChangeEvent* ev = &e->xfocus;
  if(ev->mode == NotifyGrab ||
     ev->mode == NotifyUngrab ||
     ev->detail == NotifyPointer ||
     ev->detail == NotifyInferior){
    //say("ignoring fake focus, this is wrongness\n");
    return;//ignoring
  }

  Client* c = win_to_client(MONITOR,ev->window);
  if(!c)return;
  xorg_grabbuttons(c,True);
  atomic_setprop_window(XSTATE.root,net_atoms[NetActiveWindow],c->win);
}

void event_client_focusout(XEvent* e){
  XFocusChangeEvent* ev = &e->xfocus;
  if(ev->mode == NotifyGrab ||
     ev->detail == NotifyPointer ||
     ev->detail == NotifyInferior)return;//ignoring

  Client* c = win_to_client(MONITOR,ev->window);
  if(!c)return;
  xorg_grabbuttons(c,False);
  atomic_setprop_window(XSTATE.root,net_atoms[NetActiveWindow],None);
}

void event_client_maprequest(XEvent* e) {
  //say("event: MapRequest\n");
  XWindowAttributes wa;
  XMapRequestEvent* ev = &e->xmaprequest;
  if(!XGetWindowAttributes(XSTATE.dpy, ev->window, &wa))return;
  if(wa.override_redirect)return;
  if(!win_to_client(MONITOR,ev->window)){
    Client* c=new_client(ev->window);
    c->x=wa.x;
    c->y=wa.y;
    c->w=wa.width;
    c->h=wa.height;
    xclient_manage(c);
  }
}

void event_mappingnotify(XEvent* e) {
  XMappingEvent *ev = &e->xmapping;
  XRefreshKeyboardMapping(ev);
  if(ev->request == MappingKeyboard){
    xorg_grabkeys();
  }
}

void event_configure_notify(XEvent* e) {
  XConfigureEvent* ev = &e->xconfigure;

  if(ev->window == XSTATE.root) {
    MONITOR->mw=ev->width;
    MONITOR->mh=ev->height;

    long data[2]={MONITOR->mw,MONITOR->mh};
    atomic_setprop_longa(XSTATE.root,net_atoms[NetDesktopGeometry],data,2);
    xorg_recalc();
  }
}

void event_client_message(XEvent* e){
  //say("Client Message\n");
  XClientMessageEvent* cme=&e->xclient;
  if(cme->message_type==net_atoms[NetCurrentDesktop]){
    int i=cme->data.l[0];
    user_desktopswitch(i+1);
    return;
  }
  if(cme->message_type==net_atoms[NetCloseWindow]){
    Client* c=win_to_client(MONITOR,cme->window);
    if(!c)return;
    xclient_kill(c);
    return;
  }
  if(cme->message_type==net_atoms[NetActiveWindow]){
    Client* c=win_to_client(MONITOR,cme->window);
    if(!c)return;
    if(!c->layer->visible){
      //try toggling if you don't want this
      user_togglelayer(layer_to_id(MONITOR,c->layer));
    }/*else{
      if(c==MONITOR->focus){//already focused
	//try toggling the layer visibility
	user_togglelayer(layer_to_id(MONITOR,c->layer));
      }
    }
     */
    xclient_activate(c);
    return;
  }
  if(cme->message_type==net_atoms[NetWMState]){
    Client* c=win_to_client(MONITOR,cme->window);
    if(!c)return;
    if(cme->data.l[1]==net_atoms[NetWMFullscreen]){
      int newmode=FloatFloat;
      switch(cme->data.l[0]){
	case _NET_WM_STATE_REMOVE:
	  newmode=FloatFloat;
	  break;
	case _NET_WM_STATE_ADD:
	  newmode=FloatFull;
	  break;
	case _NET_WM_STATE_TOGGLE:
	  newmode=c->float_state==FloatFull?FloatFloat:FloatFull;
	  break;
      }
      xclient_floatrel(c,newmode);
      //say("client event forced to mode %d\n",newmode);
      return;
    }
    if(cme->data.l[1]==net_atoms[NetWMMaximizedVert] ||
       cme->data.l[1]==net_atoms[NetWMMaximizedHorz]){
      int newmode=FloatFloat;
      switch(cme->data.l[0]){
	case _NET_WM_STATE_REMOVE:
	  newmode=FloatFloat;
	  break;
	case _NET_WM_STATE_ADD:
	  newmode=FloatMax;
	  break;
	case _NET_WM_STATE_TOGGLE:
	  newmode=c->float_state==FloatMax?FloatFloat:FloatMax;
	  break;
      }
      xclient_floatrel(c,newmode);
      //say("client event forced to mode %d\n",newmode);
      return;
    }
    return;
  }
}

void event_client_propertynotify(XEvent* e) {
  //say("event: PropertyNotify\n");
  XPropertyEvent* ev = &e->xproperty;
  Client *c;

  if((ev->window == XSTATE.root) && (ev->atom == XA_WM_NAME)){
    return;
    //  }else if(ev->state == PropertyDelete){
    //    return; // ignore
  }else if((c = win_to_client(MONITOR,ev->window))) {
    switch (ev->atom) {
      case XA_WM_TRANSIENT_FOR: {
	break;
	if(ev->state==PropertyDelete){
	  client_remtransient(c);
	}else{
	  Window trans;
	  XGetTransientForHint(XSTATE.dpy, c->win, &trans);
	  Client* tr=win_to_client(MONITOR,trans);
	  if(c->owner==tr)break;//we already know this
	  if(c->owner)client_remtransient(c);
	  if(!tr)break;
	  client_addtransient(tr,c);
	  if(tr->tile!=c->tile){
	    client_move_to_tile(tr->tile,c);
	  }
	}
	client_reshape(c,0);
	xclient_moveresize(c);
	break;
      }
      case XA_WM_NORMAL_HINTS:
	xorg_getsizehints(c);
	client_reshape(c,0);
	xclient_moveresize(c);
	break;
	/*
      case XA_WM_HINTS:
	xorg_updatewmhints(c);
	break;
	*/
      default: break;
    }
    if(ev->atom == XA_WM_NAME
       || ev->atom == net_atoms[NetWMName] 
       || ev->atom == wm_atoms[WMWindowRole]
       || ev->atom == wm_atoms[WMClass]){
      xorg_getstrings(c);
      return;
    }
    if(ev->atom == net_atoms[NetWMWindowType]){
      say("notify window type property 0x%x, %s\n",c->win,c->name);
      xorg_acknowledgetype(c);
      client_reshape(c,0);
      xorg_restack();
      return;
    }
  }
}

void event_configure_request(XEvent* e) {
  XConfigureRequestEvent* ev = &e->xconfigurerequest;

  Client* c=win_to_client(MONITOR,ev->window);
  int mask=0;
  if(c) {
    if(ev->value_mask & CWBorderWidth)
      //c->bw = ev->border_width;
    mask|=CWBorderWidth;
    if(ev->value_mask & CWX)
      c->x = ev->x;
    if(ev->value_mask & CWY)
      c->y = ev->y;
    if(ev->value_mask & CWWidth)
      c->w = ev->width;
    if(ev->value_mask & CWHeight)
      c->h = ev->height;
    if(ev->value_mask & (CWX|CWY|CWWidth|CWHeight)){
      client_reshape(c,0);
    }
    xclient_configure(c,(CWX|CWY|CWWidth|CWHeight|mask));
    //say("client %s configured\n",c->name);
  } else {
    XWindowChanges wc;
    wc.x = ev->x;
    wc.y = ev->y;
    wc.width = ev->width;
    wc.height = ev->height;
    wc.border_width = ev->border_width;
    wc.sibling = ev->above;
    wc.stack_mode = ev->detail;
    XConfigureWindow(XSTATE.dpy, ev->window, ev->value_mask, &wc);
    //say("window %x changed border %d, %d\n",ev->window,ev->border_width,((ev->value_mask & CWBorderWidth) !=0) );
  }
  XSync(XSTATE.dpy, False);
}


////////////////////////////////////////
void user_alttab(int arg){
  Client* f=0;
  if(!MONITOR->focus){
    for(Client* c=MONITOR->firstfocus;c;c=c->fnext){
      if(c->visible){
	f=c;
	break;
      }
    }
  }else{
    f=client_recycle(MONITOR->focus,arg);
  }
  if(!f)return;//nothing is left

  if(f!=MONITOR->focus){
    //decolorize the focus
    if(MONITOR->focus)XSetWindowBorder(XSTATE.dpy, MONITOR->focus->win, color_nb);
  }
  XRaiseWindow(XSTATE.dpy,f->win);//just raise, nothing else
  //colorize active f
  XSetWindowBorder(XSTATE.dpy, f->win, color_sb);

  XEvent ev;

  if(XGrabKeyboard(XSTATE.dpy,XSTATE.root,False,GrabModeSync,GrabModeAsync,CurrentTime)!=GrabSuccess)return;
  int carryon=True;
  do {
    XMaskEvent(XSTATE.dpy,KeyPressMask|KeyReleaseMask,&ev);
    XKeyEvent* kev=&ev.xkey;
    //deprecation
    //    KeySym keysym = XKeycodeToKeysym(XSTATE.dpy,(KeyCode)kev->keycode,0);
    KeySym keysym = XkbKeycodeToKeysym(XSTATE.dpy,(KeyCode)kev->keycode,0,0);
    if(CLEANMASK(kev->state)!=lastmask){//if mask is wrong, return
      carryon=False;break;
    }
    switch(kev->type){
      case KeyRelease:{
	//if released keycode is in the mask, return
	unsigned int keymask=xorg_keycodetomodmask(kev->keycode);
	if(keymask & lastmask)carryon=False;
	break;
      }
      case KeyPress:
	if(keysym!=XK_Tab)break;
	if(keysym==XK_Escape){
	  XSetWindowBorder(XSTATE.dpy, f->win, color_nb);
	  f=MONITOR->focus;
	  carryon=False;
	  break;
	}
	XSetWindowBorder(XSTATE.dpy, f->win, color_nb);
	f=client_recycle(f,arg);
	if(!f)break;
	XRaiseWindow(XSTATE.dpy,f->win);//just raise, nothing else
	XSetWindowBorder(XSTATE.dpy, f->win, color_sb);
	break;
      default:
	carryon=False;
    }
  }while(carryon);
  XUngrabKeyboard(XSTATE.dpy,CurrentTime);
  xclient_activate(f);//properly bring to front and restack
}

//these two do not refresh focus, don't call them anywhere.
int xlayer_bringfront(Layer* l){
  if(!l->sprev)return 0;
  if(MONITOR->sfirstlayer->focusorhide){
    xlayer_visible(MONITOR->sfirstlayer,0);
  }
  layer_stack_to_front(l);
  if(l->visible)xorg_restack();
  return 0;
}

int xlayer_bringback(Layer* l){
  if(!l->snext)return 0;
  if(MONITOR->sfirstlayer->focusorhide){
    xlayer_visible(l,0);
  }
  layer_stack_to_back(l);
  if(l->visible)xorg_restack();
  return 0;
}


void user_togglelayer(int i){
  if(MONITOR->desktop && 
     (MONITOR->desktop->layout[i]=='0'
      || MONITOR->desktop->layout[i]=='X'))return;//fixed
  Layer* l=id_to_layer(MONITOR,i);
  if(!l)return;
  xlayer_visible(l,!l->visible);
  if(l->visible){
    xlayer_bringfront(l);

    for(Client* c=MONITOR->firstfocus;c;c=c->fnext){
      if(c->layer==l){
	//prepare for focus
	client_focus_to_front(c);
	break;
      }
    }
  }else{
    xlayer_bringback(l);
  }
  int werefullscreened=MONITOR->focus && MONITOR->focus->float_state==FloatFull;
  xorg_refreshfocus();
  if(werefullscreened)xorg_restack();
}

void user_sendtolayer(int i){
  Layer* l=id_to_layer(MONITOR,i);
  if(!l)return;
  if(!MONITOR->focus)return;
  xclient_changelayer(MONITOR->focus,l);
  xorg_refreshfocus();
}

void user_sendtonexttile(int i){
  if(!MONITOR->focus)return;
  Tile* t=MONITOR->focus->tile;
  do { t=tile_cycle(t); } while(t!=MONITOR->focus->tile && !tile_is_free(t));
  if(t==MONITOR->focus->tile)return;
  client_move_to_tile(t,MONITOR->focus);
  rect_findbest(MONITOR->focus,MONITOR->focus->tile);
  client_reshape(MONITOR->focus,0);
  xclient_moveresize(MONITOR->focus);
  xclient_configure(MONITOR->focus,CWBorderWidth);
  xorg_restack();
}

void user_nextintile(int i){
  if(!MONITOR->focus)return;
  if(MONITOR->focus->tile->clientcount==1)return;
  Client* c=MONITOR->focus;
  if(i>0){
    while(i--){
      c=client_cycle(c,CycleTileFwd);
    }
  }else{
    while(i++){
      c=client_cycle(c,CycleTileBwd);
    }
  }
  xclient_activate(c);
  xorg_restack();
}


int xdesktop_setdesktop(Dsktp* d){
  if(!d)return 1;
  MONITOR->desktop=d;
  xlayer_applylayout(MONITOR->desktop->layout,1);
  atomic_setprop_long(XSTATE.root,net_atoms[NetCurrentDesktop],MONITOR->desktop->id);
  xorg_refreshalldesktop();
  return 0;
}

void user_desktopnext(int i){
  if(!i)return;
  Dsktp* d=MONITOR->desktop;
  if(!d)return;
  if(i>0){
    for(int a=0;a<i;a++){
      if(d->next){
	d=d->next;
      }else{
	d=MONITOR->firstdesktop;
      }
    }
  }else{
    for(int a=i;a;a++){
      if(d->prev){
	d=d->prev;
      }else{
	d=MONITOR->lastdesktop;
      }
    }
  }
  xdesktop_setdesktop(d);
}

void user_desktopswitch(int i){
  if(!MONITOR->desktop)return;
  i--;
  if(i==MONITOR->desktop->id)return;
  while(i<0)i+=MONITOR->desktopcount;
  while(i>=MONITOR->desktopcount)i-=MONITOR->desktopcount;
  Dsktp* newdesk=id_to_desktop(MONITOR,i);
  xdesktop_setdesktop(newdesk);
}

void user_kill(int arg){
  if(!MONITOR->focus)return;
  if(arg==0){
    xclient_kill(MONITOR->focus);
  }else{
    xclient_murder(MONITOR->focus,arg);
  }
}

void user_client_tog_flrel(int floatrel){
  if(!MONITOR->focus)return;
  if(floatrel==FloatFloat){
    if(floatrel==MONITOR->focus->float_state)return;
    xclient_floatrel(MONITOR->focus,floatrel);
    return;
  }
  if(floatrel==MONITOR->focus->float_state){
    xclient_floatrel(MONITOR->focus,FloatFloat);
    return;
  }else{
    xclient_floatrel(MONITOR->focus,floatrel);
    return;
  }
}

void user_client_tog_strel(int stackrel){
  if(!MONITOR->focus)return;
  if(MONITOR->focus->stacktype<=StckDesktop || MONITOR->focus->stacktype>=StckDock)return;
  if(stackrel==StckNormal){
    if(stackrel==MONITOR->focus->stacktype)return;
    xclient_stackrel(MONITOR->focus,stackrel);
    return;
  }
  if(stackrel==MONITOR->focus->stacktype){
    xclient_stackrel(MONITOR->focus,StckNormal);
    return;
  }else{
    xclient_stackrel(MONITOR->focus,stackrel);
    return;
  }
}


void user_quit(int arg){
  running=False;
}


void user_terminal(int arg){
  xorg_exec("urxvt",0);
}

void xorg_exec(const char* cmd, char** argv){
  if(fork() == 0){
    if(XSTATE.dpy)
      close(ConnectionNumber(XSTATE.dpy));
    setsid();
    execvp(cmd,argv);
    say("failed to execute %s");
    exit(0);
  }
}

int main(int argc, char** argv) {
  if(argc>1){
    if(!strcmp(argv[1],"-nk")){
      ignore_keys=1;
    }
  }

  if(!(XSTATE.dpy = XOpenDisplay(NULL))){
    die("wingedmonkey: cannot open display\n");
  }
  if(xorg_checkotherwm()){
    die("wingedmonkey: another wm already running\n");
  }
  char* home=getenv("HOME");
  if(!home){
    say("you are homeless, this can't be good\n");
    chdir("/tmp/");
  }else{
    chdir(home);
  }
  xorg_setup();
  xorg_scan();
  xorg_run();
  xorg_dispose();
  sigchld(0);
  XCloseDisplay(XSTATE.dpy);
  sleep(2);//wait so that the children have time to die
  return 0;
}
